__all__ = ["NetworkTable"]

from typing import Callable, List, Optional, Sequence

from ._impl.constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
    NT_PERSISTENT,
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_NEW,
    NT_NOTIFY_DELETE,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS,
)

from ._impl.value import Value

from .entry import NetworkTableEntry

import logging

logger = logging.getLogger("nt")
_is_new = NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW


class NetworkTable:
    """
        This is a NetworkTable object, it allows you to interact with
        NetworkTables in a table-based manner. You should not directly
        create a NetworkTable object, but instead use
        :meth:`.NetworkTables.getTable` to retrieve a NetworkTable instance.

        For example, to interact with the SmartDashboard::

            from networktables import NetworkTables
            sd = NetworkTables.getTable('SmartDashboard')

            someNumberEntry = sd.getEntry('someNumber')
            someNumberEntry.putNumber(1234)
            ...

        .. seealso::
           - The examples in the documentation.
           - :class:`.NetworkTablesInstance`
    """

    PATH_SEPARATOR = "/"

    def __init__(self, path, api, inst):

        #: Path of table without trailing slash
        self.path = path
        self._path = path + self.PATH_SEPARATOR
        self._pathsz = len(self._path)

        self._api = api
        self._inst = inst

        self._listeners = {}

    def __str__(self):
        return "NetworkTable: %s" % self._path

    def __repr__(self):
        return "<NetworkTable path=%s>" % self._path

    def getEntry(self, key: str) -> NetworkTableEntry:
        """Gets the entry for a subkey. This is the preferred API to use
        to access NetworkTable keys.

        :rtype: :class:`.NetworkTableEntry`

        .. versionadded:: 2018.0.0
        """
        return self._inst.getEntry(self._path + key)

    def getPath(self) -> str:
        """Gets the full path of this table.  Does not include the trailing "/".

        :returns: The path (e.g "", "/foo").
        """
        return self._path

    def addEntryListener(
        self,
        listener: Callable,
        immediateNotify: bool = False,
        key: Optional[str] = None,
        localNotify: bool = False,
    ) -> None:
        """Adds a listener that will be notified when any key in this
        NetworkTable is changed, or when a specified key changes.

        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        :param listener: A callable with signature `callable(source, key, value, isNew)`
        :param immediateNotify: If True, the listener will be called immediately with the current values of the table
        :param key: If specified, the listener will only be called when this key is changed
        :param localNotify: True if you wish to be notified of changes made locally (default is False)

        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended

        .. versionchanged:: 2017.0.0
           Added localNotify parameter (defaults to False, which is different from NT2)

        """
        flags = NT_NOTIFY_NEW | NT_NOTIFY_UPDATE
        if immediateNotify:
            flags |= NT_NOTIFY_IMMEDIATE
        if localNotify:
            flags |= NT_NOTIFY_LOCAL

        self.addEntryListenerEx(listener, flags, key=key)

    def addEntryListenerEx(
        self,
        listener: Callable,
        flags: int,
        key: Optional[str] = None,
        paramIsNew: bool = True,
    ) -> None:
        """Adds a listener that will be notified when any key in this
        NetworkTable is changed, or when a specified key changes.

        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        :param listener: A callable with signature `callable(source, key, value, param)`
        :param flags: Bitmask of flags that indicate the types of notifications you wish to receive
        :type flags: :class:`.NotifyFlags`
        :param key: If specified, the listener will only be called when this key is changed
        :param paramIsNew: If True, the listener fourth parameter is a boolean set to True
                           if the listener is being called because of a new value in the
                           table. Otherwise, the parameter is an integer of the raw
                           `NT_NOTIFY_*` flags

        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended

        .. versionadded:: 2017.0.0
        """

        # if key is None:
        # Any key in this table (but not subtables)

        if key is None:

            # Any key in this table (but not subtables)
            _pathsz = self._pathsz
            if paramIsNew:

                def callback(item):
                    key_, value_, flags_, _ = item
                    key_ = key_[_pathsz:]
                    if "/" not in key_:
                        listener(self, key_, value_.value, (flags_ & _is_new) != 0)

            else:

                def callback(item):
                    key_, value_, flags_, _ = item
                    key_ = key_[_pathsz:]
                    if "/" not in key_:
                        listener(self, key_, value_.value, flags_)

            uid = self._api.addEntryListener(self._path, callback, flags)

        # Hack: Internal flag used by addGlobalListener*
        elif key == 0xDEADBEEF:
            if paramIsNew:

                def callback(item):
                    key_, value_, flags_, _ = item
                    listener(key_, value_.value, (flags_ & _is_new) != 0)

            else:
                callback = listener

            uid = self._api.addEntryListener("/", callback, flags)
        else:
            entry_id = self._api.getEntryId(self._path + key)
            uid = self._api.addEntryListenerByIdEx(
                self, key, entry_id, listener, flags, paramIsNew
            )

        self._listeners.setdefault(listener, []).append(uid)

    # deprecated aliases
    addTableListener = addEntryListener
    addTableListenerEx = addEntryListenerEx

    def addSubTableListener(
        self, listener: Callable, localNotify: bool = False
    ) -> None:
        """Adds a listener that will be notified when any key in a subtable of
        this NetworkTable is changed.

        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        :param listener: Callable to call when previously unseen table appears.
                         Function signature is `callable(source, key, subtable, True)`
        :param localNotify: True if you wish to be notified when local changes
                            result in a new table

        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended as we are not
                     currently sure if deadlocks will occur

        .. versionchanged:: 2017.0.0
           Added localNotify parameter
        """
        notified_tables = {}

        def _callback(item):
            key, value_, _1, _2 = item
            key = key[self._pathsz :]
            if "/" in key:
                skey = key[: key.index("/")]

                o = object()
                if notified_tables.setdefault(skey, o) is o:
                    try:
                        listener(self, skey, self.getSubTable(skey), True)
                    except Exception:
                        logger.warning(
                            "Unhandled exception in %s", listener, exc_info=True
                        )

        flags = NT_NOTIFY_NEW | NT_NOTIFY_IMMEDIATE
        if localNotify:
            flags |= NT_NOTIFY_LOCAL

        uid = self._api.addEntryListener(self._path, _callback, flags)
        self._listeners.setdefault(listener, []).append(uid)

    def removeEntryListener(self, listener: Callable) -> None:
        """Removes a table listener

        :param listener: callable that was passed to :meth:`.addTableListener`
                         or :meth:`.addSubTableListener`
        """
        uids = self._listeners.pop(listener, [])
        for uid in uids:
            self._api.removeEntryListener(uid)

    # Deprecated alias
    removeTableListener = removeEntryListener

    def getSubTable(self, key: str) -> "NetworkTable":
        """Returns the table at the specified key. If there is no table at the
        specified key, it will create a new table

        :param key: the key name

        :returns: the networktable to be returned
        """
        path = self._path + key
        return self._inst.getTable(path)

    def containsKey(self, key: str) -> bool:
        """Determines whether the given key is in this table.

        :param key: the key to search for

        :returns: True if the table as a value assigned to the given key
        """
        path = self._path + key
        return self._api.getEntryValue(path) is not None

    def __contains__(self, key: str) -> bool:
        return self.containsKey(key)

    def containsSubTable(self, key: str) -> bool:
        """Determines whether there exists a non-empty subtable for this key
        in this table.

        :param key: the key to search for (must not end with path separator)

        :returns: True if there is a subtable with the key which contains at least
                  one key/subtable of its own
        """
        path = self._path + key + self.PATH_SEPARATOR
        return len(self._api.getEntryInfo(path, 0)) > 0

    def getKeys(self, types: int = 0) -> List[str]:
        """
        :param types: bitmask of types; 0 is treated as a "don't care".
        :type types: :class:`.EntryTypes`

        :returns: keys currently in the table

        .. versionadded:: 2017.0.0
        """
        keys = []
        for entry in self._api.getEntryInfo(self._path, types):
            relative_key = entry.name[len(self._path) :]
            if self.PATH_SEPARATOR in relative_key:
                continue

            keys.append(relative_key)

        return keys

    def getSubTables(self) -> List[str]:
        """:returns: subtables currently in the table

        .. versionadded:: 2017.0.0
        """
        keys = set()
        for entry in self._api.getEntryInfo(self._path, 0):
            relative_key = entry.name[len(self._path) :]
            subst = relative_key.split(self.PATH_SEPARATOR)
            if len(subst) == 1:
                continue

            keys.add(subst[0])

        return list(keys)

    def setPersistent(self, key: str) -> None:
        """Makes a key's value persistent through program restarts.

        :param key: the key to make persistent

        .. versionadded:: 2017.0.0
        """
        self.setFlags(key, NT_PERSISTENT)

    def clearPersistent(self, key: str) -> None:
        """Stop making a key's value persistent through program restarts.

        :param key: the key name

        .. versionadded:: 2017.0.0
        """
        self.clearFlags(key, NT_PERSISTENT)

    def isPersistent(self, key: str) -> bool:
        """Returns whether the value is persistent through program restarts.

        :param key: the key name

        .. versionadded:: 2017.0.0
        """
        return self.getFlags(key) & NT_PERSISTENT != 0

    def delete(self, key: str) -> None:
        """Deletes the specified key in this table.

        :param key: the key name

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        self._api.deleteEntry(path)

    def setFlags(self, key: str, flags: int) -> None:
        """Sets entry flags on the specified key in this table.

        :param key: the key name
        :param flags: the flags to set (bitmask)
        :type flags: :class:`.EntryFlags`

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        self._api.setEntryFlags(path, self._api.getEntryFlags(path) | flags)

    def clearFlags(self, key: str, flags: int) -> None:
        """Clears entry flags on the specified key in this table.

        :param key: the key name
        :param flags: the flags to clear (bitmask)
        :type flags: :class:`.EntryFlags`

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        self._api.setEntryFlags(path, self._api.getEntryFlags(path) & ~flags)

    def getFlags(self, key: str):
        """Returns the entry flags for the specified key.

        :param key: the key name
        :returns: the flags, or 0 if the key is not defined
        :rtype: :class:`.EntryFlags`

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.getEntryFlags(path)

    def putNumber(self, key: str, value: float) -> bool:
        """Put a number in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeDouble(value))

    def setDefaultNumber(self, key: str, defaultValue: float) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: int, float

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeDouble(defaultValue))

    def getNumber(self, key: str, defaultValue: float) -> float:
        """Gets the number associated with the given name.

        :param key: the key to look up
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_DOUBLE:
            return defaultValue

        return value.value

    def putString(self, key: str, value: str) -> bool:
        """Put a string in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeString(value))

    def setDefaultString(self, key: str, defaultValue: str) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeString(defaultValue))

    def getString(self, key: str, defaultValue: str) -> str:
        """Gets the string associated with the given name. If the key does not
        exist or is of different type, it will return the default value.

        :param key: the key to look up
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_STRING:
            return defaultValue

        return value.value

    def putBoolean(self, key: str, value: bool) -> bool:
        """Put a boolean in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeBoolean(value))

    def setDefaultBoolean(self, key: str, defaultValue: bool) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeBoolean(defaultValue))

    def getBoolean(self, key: str, defaultValue: bool) -> bool:
        """Gets the boolean associated with the given name. If the key does not
        exist or is of different type, it will return the default value.

        :param key: the key name
        :param defaultValue: the default value if no value is found

        :returns: the key
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_BOOLEAN:
            return defaultValue

        return value.value

    def putBooleanArray(self, key: str, value: Sequence[bool]) -> bool:
        """Put a boolean array in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeBooleanArray(value))

    def setDefaultBooleanArray(self, key: str, defaultValue: Sequence[bool]) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(
            path, Value.makeBooleanArray(defaultValue)
        )

    def getBooleanArray(self, key: str, defaultValue) -> Sequence[bool]:
        """Returns the boolean array the key maps to. If the key does not exist or is
        of different type, it will return the default value.

        :param key: the key to look up
        :type key: str
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_BOOLEAN_ARRAY:
            return defaultValue

        return value.value

    def putNumberArray(self, key: str, value: Sequence[float]) -> bool:
        """Put a number array in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeDoubleArray(value))

    def setDefaultNumberArray(self, key: str, defaultValue: Sequence[float]) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeDoubleArray(defaultValue))

    def getNumberArray(self, key: str, defaultValue) -> Sequence[float]:
        """Returns the number array the key maps to. If the key does not exist or is
        of different type, it will return the default value.

        :param key: the key to look up
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_DOUBLE_ARRAY:
            return defaultValue

        return value.value

    def putStringArray(self, key: str, value: Sequence[str]) -> bool:
        """Put a string array in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeStringArray(value))

    def setDefaultStringArray(self, key: str, defaultValue: Sequence[str]) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeStringArray(defaultValue))

    def getStringArray(self, key: str, defaultValue) -> Sequence[str]:
        """Returns the string array the key maps to. If the key does not exist or is
        of different type, it will return the default value.

        :param key: the key to look up
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_STRING_ARRAY:
            return defaultValue

        return value.value

    def putRaw(self, key: str, value: bytes) -> bool:
        """Put a raw value (byte array) in the table

        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setEntryValue(path, Value.makeRaw(value))

    def setDefaultRaw(self, key: str, defaultValue: bytes) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        return self._api.setDefaultEntryValue(path, Value.makeRaw(defaultValue))

    def getRaw(self, key: str, defaultValue: bytes) -> bytes:
        """Returns the raw value (byte array) the key maps to. If the key does not
        exist or is of different type, it will return the default value.

        :param key: the key to look up
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value or value.type != NT_RAW:
            return defaultValue

        return value.value

    def putValue(self, key: str, value) -> bool:
        """Put a value in the table, trying to autodetect the NT type of
        the value. Refer to this table to determine the type mapping:

        ======= ============================ =================================
        PyType  NT Type                       Notes
        ======= ============================ =================================
        bool    :attr:`.EntryTypes.BOOLEAN`
        int     :attr:`.EntryTypes.DOUBLE`
        float   :attr:`.EntryTypes.DOUBLE`
        str     :attr:`.EntryTypes.STRING`
        bytes   :attr:`.EntryTypes.RAW`
        list    Error                        Use `putXXXArray` methods instead
        tuple   Error                        Use `putXXXArray` methods instead
        ======= ============================ =================================

        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :type value: bool, int, float, str, bytes

        :returns: False if the table key already exists with a different type

        .. versionadded:: 2017.0.0
        """
        value = Value.getFactory(value)(value)
        path = self._path + key
        return self._api.setEntryValue(path, value)

    def setDefaultValue(self, key: str, defaultValue) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.

        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.
        :type defaultValue: bool, int, float, str, bytes

        :returns: False if the table key exists with a different type

        .. versionadded:: 2017.0.0

        .. seealso:: :meth:`.putValue`
        """
        defaultValue = Value.getFactory(defaultValue)(defaultValue)
        path = self._path + key
        return self._api.setDefaultEntryValue(path, defaultValue)

    def getValue(self, key: str, defaultValue):
        """Gets the value associated with a key. This supports all
        NetworkTables types (unlike :meth:`putValue`).

        :param key: the key of the value to look up
        :param defaultValue: The default value to return if the key doesn't exist
        :type defaultValue: any

        :returns: the value associated with the given key
        :rtype: bool, int, float, str, bytes, tuple

        .. versionadded:: 2017.0.0
        """
        path = self._path + key
        value = self._api.getEntryValue(path)
        if not value:
            return defaultValue

        return value.value

    def getAutoUpdateValue(
        self, key: str, defaultValue, writeDefault: bool = True
    ) -> NetworkTableEntry:
        """Returns an object that will be automatically updated when the
        value is updated by networktables.

        :param key: the key name
        :param defaultValue: Default value to use if not in the table
        :type  defaultValue: any
        :param writeDefault: If True, put the default value to the table,
                             overwriting existing values

        :rtype: :class:`.NetworkTableEntry`

        .. note:: If you modify the returned value, the value will NOT
                  be written back to NetworkTables (though now there are functions
                  you can use to write values). See :func:`.ntproperty` if
                  you're looking for that sort of thing.

        .. seealso:: :func:`.ntproperty` is a better alternative to use

        .. versionadded:: 2015.1.3

        .. versionchanged:: 2018.0.0
           This now returns the same as :meth:`.NetworkTable.getEntry`

        """
        return self._inst.getGlobalAutoUpdateValue(
            self._path + key, defaultValue, writeDefault
        )



from _pynetworktables import (
    NetworkTablesInstance as NetworkTablesInstance,
    NetworkTables as NetworkTables,
    NetworkTable as NetworkTable,
    NetworkTableEntry as NetworkTableEntry,
    Value as Value,
    __version__ as __version__,
)

nt_backend: str = ...

__all__ = (
    "NetworkTablesInstance",
    "NetworkTables",
    "NetworkTable",
    "NetworkTableEntry",
    "Value",
)



# todo: tracks NetworkTablesInstance.java

from typing import Any, Callable, List, Optional, Sequence, Tuple, Union
from weakref import WeakSet

from ._impl import constants
from ._impl.api import NtCoreApi

from .entry import NetworkTableEntry
from .table import NetworkTable

import logging

logger = logging.getLogger("nt")

__all__ = ["NetworkTablesInstance"]

ServerPortPair = Tuple[str, int]


class NetworkTablesInstance:
    """
        The object ``networktables.NetworkTables`` is a global singleton that
        you can use to initialize NetworkTables connections, configure global
        settings and listeners, and to create table objects which can be used
        to send data to/from NetworkTable servers and clients.

        First, you must initialize NetworkTables::

            from networktables import NetworkTables

            # As a client to connect to a robot
            NetworkTables.initialize(server='roborio-XXX-frc.local')

        Then, to interact with the SmartDashboard you get an instance of the
        table, and you can call the various methods::

            sd = NetworkTables.getTable('SmartDashboard')

            sd.putNumber('someNumber', 1234)
            otherNumber = sd.getNumber('otherNumber')

        You can create additional NetworkTablesInstance objects.
        Instances are completely independent from each other.  Table operations on
        one instance will not be visible to other instances unless the instances are
        connected via the network.  The main limitation on instances is that you
        cannot have two servers on the same network port.  The main utility of
        instances is for unit testing, but they can also enable one program to
        connect to two different NetworkTables networks.

        The global "default" instance (as returned by :meth:`.NetworkTablesInstance.getDefault`) is
        always available, and is intended for the common case when there is only
        a single NetworkTables instance being used in the program.

        Additional instances can be created with the :meth:`.create` function.

        .. seealso::
           - The examples in the documentation.
           - :class:`.NetworkTable`
    """

    class EntryTypes:
        """
            NetworkTable value types used in :meth:`.NetworkTable.getKeys`
        """

        #: True or False
        BOOLEAN = constants.NT_BOOLEAN

        #: Floating point number
        DOUBLE = constants.NT_DOUBLE

        #: Strings
        STRING = constants.NT_STRING

        #: Raw bytes
        RAW = constants.NT_RAW

        #: List of booleans
        BOOLEAN_ARRAY = constants.NT_BOOLEAN_ARRAY

        #: List of numbers
        DOUBLE_ARRAY = constants.NT_DOUBLE_ARRAY

        #: List of strings
        STRING_ARRAY = constants.NT_STRING_ARRAY

    class EntryFlags:
        """
            NetworkTables entry flags
        """

        #: Indicates a value that will be persisted on the server
        PERSISTENT = constants.NT_PERSISTENT

    class NotifyFlags:
        """
            Bitflags passed to entry callbacks
        """

        #: Initial listener addition
        IMMEDIATE = constants.NT_NOTIFY_IMMEDIATE

        #: Changed locally
        LOCAL = constants.NT_NOTIFY_LOCAL

        #: Newly created entry
        NEW = constants.NT_NOTIFY_NEW

        #: Key deleted
        DELETE = constants.NT_NOTIFY_DELETE

        #: Value changed
        UPDATE = constants.NT_NOTIFY_UPDATE

        #: Flags changed
        FLAGS = constants.NT_NOTIFY_FLAGS

    class NetworkModes:
        """
            Bitflags returend from :meth:`.getNetworkMode`
        """

        #: Not running
        NONE = constants.NT_NET_MODE_NONE

        #: Running in server mode
        SERVER = constants.NT_NET_MODE_SERVER

        #: Running in client mode
        CLIENT = constants.NT_NET_MODE_CLIENT

        #: Flag for starting (either client or server)
        STARTING = constants.NT_NET_MODE_STARTING

        #: Flag for failure (either client or server)
        FAILURE = constants.NT_NET_MODE_FAILURE

        #: Flag indicating in test mode
        TEST = constants.NT_NET_MODE_TEST

    #: The path separator for sub-tables and keys
    PATH_SEPARATOR = "/"

    #: The default port that network tables operates on
    DEFAULT_PORT = constants.NT_DEFAULT_PORT

    @classmethod
    def create(cls) -> "NetworkTablesInstance":
        """Create an instance.

        :returns: Newly created instance
        """
        return cls()

    @classmethod
    def getDefault(cls) -> "NetworkTablesInstance":
        """Get global default instance."""
        try:
            return cls._defaultInstance
        except AttributeError:
            pass

        cls._defaultInstance = cls()
        return cls._defaultInstance

    def __init__(self):
        self._init()

    def _init(self):
        self._api = NtCoreApi(self.__createEntry)
        self._tables = {}
        self._entry_listeners = {}
        self._conn_listeners = {}

        if not hasattr(self, "_ntproperties"):
            self._ntproperties = WeakSet()
        else:
            for ntprop in self._ntproperties:
                ntprop.reset()

    def __createEntry(self, key, local_id):
        return NetworkTableEntry(self._api, local_id, key)

    def getEntry(self, name: str) -> NetworkTableEntry:
        """Gets the entry for a key.

        :param name: Absolute path of key
        :returns: Network table entry.

        .. versionadded:: 2018.0.0
        """
        assert name.startswith("/")
        return self._api.getEntry(name)

    def getEntries(self, prefix: str, types: int = 0) -> Sequence[NetworkTableEntry]:
        """Get entries starting with the given prefix.
        The results are optionally filtered by string prefix and entry type to
        only return a subset of all entries.

        :param prefix: entry name required prefix; only entries whose name
                       starts with this string are returned
        :param types: bitmask of types; 0 is treated as a "don't care"
        :returns: List of matching entries.
        :rtype: list of :class:`.NetworkTableEntry`

        .. versionadded:: 2018.0.0
        """
        return self._api.getEntries(prefix, types)

    def getEntryInfo(self, prefix: str, types: int = 0) -> Sequence:
        """Get information about entries starting with the given prefix.
        The results are optionally filtered by string prefix and entry type to
        only return a subset of all entries.

        :param prefix: entry name required prefix; only entries whose name
                       starts with this string are returned
        :param types: bitmask of types; 0 is treated as a "don't care"
        :returns: List of entry information.

        .. versionadded:: 2018.0.0
        """
        return self._api.getEntryInfo(prefix, types)

    def getTable(self, key: str) -> NetworkTable:
        """Gets the table with the specified key.

        :param key: the key name

        :returns: the network table requested

        .. versionchanged:: 2018.0.0
           No longer automatically initializes network tables

        """

        # Must start with separator
        if key.startswith(self.PATH_SEPARATOR):
            path = key
        else:
            path = self.PATH_SEPARATOR + key

        # Must not end with separator
        if path.endswith(self.PATH_SEPARATOR):
            path = path[:-1]

        table = self._tables.get(path)
        if table is None:
            table = NetworkTable(path, self._api, self)
            table = self._tables.setdefault(path, table)
        return table

    def deleteAllEntries(self) -> None:
        """Deletes ALL keys in ALL subtables (except persistent values).
        Use with caution!

        .. versionadded:: 2018.0.0
        """
        self._api.deleteAllEntries()

    # Deprecated alias
    globalDeleteAll = deleteAllEntries

    def addEntryListener(
        self,
        listener: Callable[[str, Any, int], None],
        immediateNotify: bool = True,
        localNotify: bool = True,
        paramIsNew: bool = True,
    ) -> None:
        """Adds a listener that will be notified when any key in any
        NetworkTable is changed. The keys that are received using this
        listener will be full NetworkTable keys. Most users will not
        want to use this listener type.

        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        :param listener: A callable that has this signature: `callable(key, value, isNew)`
        :param immediateNotify: If True, the listener will be called immediately with the current values of the table
        :param localNotify: True if you wish to be notified of changes made locally (default is True)
        :param paramIsNew: If True, the listener third parameter is a boolean set to True
                           if the listener is being called because of a new value in the
                           table. Otherwise, the parameter is an integer of the raw
                           `NT_NOTIFY_*` flags

        .. versionadded:: 2015.2.0

        .. versionchanged:: 2017.0.0
           `paramIsNew` parameter added

        .. versionchanged:: 2018.0.0
           Renamed to addEntryListener, no longer initializes NetworkTables

        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended as we are not
                     currently sure if deadlocks will occur
        """
        gtable = self.getTable("/")
        gtable.addTableListener(
            listener,
            immediateNotify=immediateNotify,
            key=0xDEADBEEF,
            localNotify=localNotify,
        )

    def addEntryListenerEx(
        self,
        listener: Callable[[str, Any, int], None],
        flags: int,
        paramIsNew: bool = True,
    ) -> None:
        """Adds a listener that will be notified when any key in any
        NetworkTable is changed. The keys that are received using this
        listener will be full NetworkTable keys. Most users will not
        want to use this listener type.

        The listener is called from the NetworkTables I/O thread, and should
        return as quickly as possible.

        :param listener: A callable that has this signature: `callable(key, value, isNew)`
        :param flags: Bitmask of flags that indicate the types of notifications you wish to receive
        :type flags: :class:`.NotifyFlags`
        :param paramIsNew: If True, the listener third parameter is a boolean set to True
                           if the listener is being called because of a new value in the
                           table. Otherwise, the parameter is an integer of the raw
                           `NT_NOTIFY_*` flags

        .. versionadded:: 2017.0.0

        .. versionchanged:: 2018.0.0
           Renamed to addEntryListenerEx, no longer initializes NetworkTables

        """
        gtable = self.getTable("/")
        gtable.addTableListenerEx(
            listener, flags, key=0xDEADBEEF, paramIsNew=paramIsNew
        )

    # deprecated aliases
    addGlobalListener = addEntryListener
    addGlobalListenerEx = addEntryListenerEx

    def removeEntryListener(self, listener: Callable[[str, Any, int], None]) -> None:
        """Remove an entry listener.

        :param listener: Listener to remove

        .. versionadded:: 2018.0.0

        """
        self.getTable("/").removeEntryListener(listener)

    # Deprecated alias
    removeGlobalListener = removeEntryListener

    def waitForEntryListenerQueue(self, timeout: float) -> bool:
        """Wait for the entry listener queue to be empty.  This is primarily useful
        for deterministic testing.  This blocks until either the entry listener
        queue is empty (e.g. there are no more events that need to be passed along
        to callbacks or poll queues) or the timeout expires.

        .. warning:: This function is not efficient, so only use it for testing!

        :param timeout: timeout, in seconds.  Set to 0 for non-blocking behavior,
                        or None to block indefinitely
        :returns: False if timed out, otherwise true.
        """
        return self._api.waitForEntryListenerQueue(timeout)

    def addConnectionListener(self, listener: Callable, immediateNotify: bool = False):
        """Adds a listener that will be notified when a new connection to a
        NetworkTables client/server is established.

        The listener is called from a NetworkTables owned thread and should
        return as quickly as possible.

        :param listener: A function that will be called with two parameters
        :type listener: fn(bool, ConnectionInfo)

        :param immediateNotify: If True, the listener will be called immediately
                                with any active connection information

        .. warning:: You may call the NetworkTables API from within the
                     listener, but it is not recommended.

        .. versionchanged:: 2017.0.0
           The listener is now a function
        """
        assert callable(listener)
        cb = lambda info: listener(info.connected, info.conn_info)
        listener_id = self._api.addConnectionListener(cb, immediateNotify)
        self._conn_listeners.setdefault(listener, []).append(listener_id)

    def removeConnectionListener(self, listener: Callable):
        """Removes a connection listener

        :param listener: The function registered for connection notifications
        """
        for listener_id in self._conn_listeners.pop(listener, []):
            self._api.removeConnectionListener(listener_id)

    def waitForConnectionListenerQueue(self, timeout: float) -> bool:
        """Wait for the connection listener queue to be empty.  This is primarily useful
        for deterministic testing.  This blocks until either the connection listener
        queue is empty (e.g. there are no more events that need to be passed along
        to callbacks or poll queues) or the timeout expires.

        :param timeout: timeout, in seconds.  Set to 0 for non-blocking behavior,
                        or a negative value to block indefinitely
        :returns: False if timed out, otherwise true.

        .. versionadded:: 2018.0.0
        """
        return self._api.waitForConnectionListenerQueue(timeout)

    #
    # Client/Server functions
    #

    def setNetworkIdentity(self, name: str) -> None:
        """Sets the network identity of this node. This is the name used in the
        initial connection handshake, and is provided in the connection info
        on the remote end.

        :param name: A string to communicate to other NetworkTables instances

        .. versionadded:: 2017.0.0
        """
        self._api.setNetworkIdentity(name)

    def getNetworkMode(self):
        """Get the current network mode

        .. versionadded:: 2018.0.0
        """
        return self._api.getNetworkMode()

    def isServer(self) -> bool:
        """:returns: True if configured in server mode"""
        return (self.getNetworkMode() & self.NetworkModes.SERVER) != 0

    def startServer(
        self,
        persistFilename: str = "networktables.ini",
        listenAddress: str = "",
        port: int = constants.NT_DEFAULT_PORT,
    ):
        """Starts a server using the specified filename, listening address, and port.

        :param persistFilename: the name of the persist file to use
        :param listenAddress: the address to listen on, or empty to listen on any
                              address
        :param port: port to communicate over

        .. versionadded:: 2018.0.0
        """
        return self._api.startServer(persistFilename, listenAddress, port)

    def stopServer(self) -> None:
        """Stops the server if it is running.

        .. versionadded:: 2018.0.0
        """
        self._api.stopServer()

    def startClient(
        self,
        server_or_servers: Union[str, ServerPortPair, List[ServerPortPair], List[str]],
    ):
        """Sets server addresses and port for client (without restarting client).
        The client will attempt to connect to each server in round robin fashion.

        :param server_or_servers: a string, a tuple of (server, port), array of
                                  (server, port), or an array of strings

        .. versionadded:: 2018.0.0
        """
        self.setServer(server_or_servers)
        return self._api.startClient()

    def startClientTeam(self, team: int, port: int = constants.NT_DEFAULT_PORT):
        """Starts a client using commonly known robot addresses for the specified
        team.

        :param team: team number
        :param port: port to communicate over

        .. versionadded:: 2018.0.0
        """
        self.setServerTeam(team, port)
        return self._api.startClient()

    def stopClient(self) -> None:
        """Stops the client if it is running.

        .. versionadded:: 2018.0.0
        """
        self._api.stopClient()

    def setServer(
        self,
        server_or_servers: Union[str, ServerPortPair, List[ServerPortPair], List[str]],
    ) -> None:
        """Sets server addresses and port for client (without restarting client).
        The client will attempt to connect to each server in round robin fashion.

        :param server_or_servers: a string, a tuple of (server, port), array of
                                  (server, port), or an array of strings

        .. versionadded:: 2018.0.0
        """
        if isinstance(server_or_servers, list):
            server_or_servers = [
                item if isinstance(item, tuple) else (item, constants.NT_DEFAULT_PORT)
                for item in server_or_servers
            ]
        elif isinstance(server_or_servers, str):
            server_or_servers = [(server_or_servers, constants.NT_DEFAULT_PORT)]

        self._api.setServer(server_or_servers)

    def setServerTeam(self, team: int, port: int = constants.NT_DEFAULT_PORT) -> None:
        """Sets server addresses and port for client based on the team number
        (without restarting client). The client will attempt to connect to each
        server in round robin fashion.

        :param team: Team number
        :param port: Port to communicate over

        .. versionadded:: 2018.0.0
        """
        self._api.setServerTeam(team, port)

    def startDSClient(self, port: int = constants.NT_DEFAULT_PORT) -> None:
        """Starts requesting server address from Driver Station.
        This connects to the Driver Station running on localhost to obtain the
        server IP address.

        :param port: server port to use in combination with IP from DS

        .. versionadded:: 2018.0.0
           Was formerly called setDashboardMode
        """
        self._api.startDSClient(port)

    setDashboardMode = startDSClient

    def setUpdateRate(self, interval: float) -> None:
        """Sets the period of time between writes to the network.

        WPILib's networktables and SmartDashboard default to 100ms, we have
        set it to 50ms instead for quicker response time. You should not set
        this value too low, as it could potentially increase the volume of
        data sent over the network.

        :param interval: Write flush period in seconds (default is 0.050,
                         or 50ms)

        .. warning:: If you don't know what this setting affects, don't mess
                     with it!

        .. versionadded:: 2017.0.0
        """
        self._api.setUpdateRate(interval)

    def flush(self) -> None:
        """Flushes all updated values immediately to the network.

        .. note:: This is rate-limited to protect the network from flooding.
                  This is primarily useful for synchronizing network updates
                  with user code.

        .. versionadded:: 2017.0.0
        """
        self._api.flush()

    def getConnections(self) -> Sequence:
        """Gets information on the currently established network connections.
        If operating as a client, this will return either zero or one values.

        :returns: list of connection information
        :rtype: list

        .. versionadded:: 2018.0.0
        """
        return self._api.getConnections()

    def getRemoteAddress(self) -> Optional[str]:
        """
            Only returns a valid address if connected to the server. If
            this is a server, returns None

            :returns: IP address of server or None

            .. versionadded:: 2015.3.2
        """
        return self._api.getRemoteAddress()

    def isConnected(self) -> bool:
        """
            :returns: True if connected to at least one other NetworkTables
                      instance
        """
        return self._api.getIsConnected()

    def savePersistent(self, filename: str):
        """Saves persistent keys to a file. The server does this automatically.

        :param filename: Name of file to save keys to

        :returns: None if success, or a string describing the error on failure

        .. versionadded:: 2017.0.0
        """
        return self._api.savePersistent(filename)

    def loadPersistent(self, filename: str):
        """Loads persistent keys from a file. WPILib will do this automatically
        on a robot server.

        :param filename: Name of file to load keys from

        :returns: None if success, or a string describing the error on failure

        .. versionadded:: 2017.0.0
        """
        return self._api.loadPersistent(filename)

    def saveEntries(self, filename: str, prefix: str):
        """Save table values to a file.  The file format used is identical to
        that used for SavePersistent.

        :param filename: filename
        :param prefix: save only keys starting with this prefix

        :returns: None if success, or a string describing the error on failure

        .. versionadded:: 2018.0.0
        """
        return self._api.saveEntries(filename, prefix)

    def loadEntries(self, filename: str, prefix: str):
        """Load table values from a file.  The file format used is identical to
        that used for SavePersistent / LoadPersistent.

        :param filename: filename
        :param prefix: load only keys starting with this prefix

        :returns: None if success, or a string describing the error on failure

        .. versionadded:: 2018.0.0
        """
        return self._api.loadEntries(filename, prefix)

    #
    # These methods are unique to pynetworktables
    #

    def initialize(self, server=None):
        """Initializes NetworkTables and begins operations

        :param server: If specified, NetworkTables will be set to client
                       mode and attempt to connect to the specified server.
                       This is equivalent to executing::

                           self.startClient(server)
        :type server: str

        :returns: True if initialized, False if already initialized

        .. versionadded:: 2017.0.0
           The *server* parameter
        """

        if server is not None:
            return self.startClient(server)
        else:
            return self.startServer()

    def shutdown(self) -> None:
        """Stops all NetworkTables activities and unregisters all tables
        and callbacks. You can call :meth:`.initialize` again after
        calling this.

        .. versionadded:: 2017.0.0
        """

        self._api.stop()
        self._api.destroy()

        self._init()

    def startTestMode(self, server: bool = True):
        """Setup network tables to run in unit test mode, and enables verbose
        logging.

        :returns: True if successful

        .. versionadded:: 2018.0.0
        """
        self.enableVerboseLogging()
        return self._api.startTestMode(server)

    def enableVerboseLogging(self) -> None:
        """Enable verbose logging that can be useful when trying to diagnose
        NetworkTables issues.

        .. warning:: Don't enable this in normal use, as it can potentially
                     cause performance issues due to the volume of logging.

        .. versionadded:: 2017.0.0
        """
        self._api.setVerboseLogging(True)

    def getGlobalTable(self) -> NetworkTable:
        """Returns an object that allows you to write values to absolute
        NetworkTable keys (which are paths with / separators).

        .. note:: This is now an alias for ``NetworkTables.getTable('/')``

        .. versionadded:: 2015.2.0
        .. versionchanged:: 2017.0.0
           Returns a NetworkTable instance
        .. versionchanged:: 2018.0.0
           No longer automatically initializes network tables
        """
        return self.getTable("/")

    def getGlobalAutoUpdateValue(
        self, key: str, defaultValue, writeDefault: bool
    ) -> NetworkTableEntry:
        """Global version of getAutoUpdateValue.

        :param key: the full NT path of the value (must start with /)
        :param defaultValue: The default value to return if the key doesn't exist
        :param writeDefault: If True, force the value to the specified default

        :rtype: :class:`.NetworkTableEntry`

        .. seealso:: :func:`.ntproperty` is a read-write alternative to this

        .. versionadded:: 2015.3.0

        .. versionchanged:: 2018.0.0
           This now returns the same as :meth:`NetworkTablesInstance.getEntry`
        """
        assert key.startswith("/")

        entry = self.getEntry(key)
        if writeDefault:
            entry.forceSetValue(defaultValue)
        else:
            entry.setDefaultValue(defaultValue)

        return entry



import threading

import logging

logger = logging.getLogger("nt.th")


class SafeThread(object):
    """
        Not exactly the same as wpiutil SafeThread... exists so we don't have
        to duplicate functionality in a lot of places
    """

    # Name each thread uniquely to make debugging easier
    _global_indices_lock = threading.Lock()
    _global_indices = {}

    def __init__(self, target, name, args=()):
        """
            Note: thread is automatically started and daemonized
        """

        with SafeThread._global_indices_lock:
            idx = SafeThread._global_indices.setdefault(name, -1) + 1
            SafeThread._global_indices[name] = idx
            name = "%s-%s" % (name, idx)

        self.name = name

        self._thread = threading.Thread(
            target=self._run, name=name, args=(target, args)
        )
        self._thread.daemon = True

        self.is_alive = self._thread.is_alive
        self.join = self._thread.join

        self._thread.start()

    def join(self, timeout=1):
        self._thread.join(timeout=timeout)
        if not self._thread.is_alive():
            logger.warning("Thread %s did not stop!", self.name)

    def _run(self, target, args):
        logger.debug("Started thread %s", self.name)
        try:
            target(*args)
        except Exception:
            logger.warning("Thread %s died unexpectedly", self.name, exc_info=True)
        else:
            logger.debug("Thread %s exited", self.name)



import warnings

from . import NetworkTableEntry  # noqa

warnings.warn(
    "networktables.entry is deprecated, import networktables.NetworkTableEntry directly",
    DeprecationWarning,
    stacklevel=2,
)



from typing import Callable, Optional, Sequence

# from pynetworktables._impl.value import Value

from . import NetworkTablesInstance, Value

__all__ = ["ntproperty", "ChooserControl"]

NetworkTables = NetworkTablesInstance.getDefault()


class _NtProperty:
    def __init__(
        self,
        key: str,
        defaultValue,
        writeDefault: bool,
        persistent: bool,
        inst: NetworkTablesInstance,
    ) -> None:
        self.key = key
        self.defaultValue = defaultValue
        self.writeDefault = writeDefault
        self.persistent = persistent
        # never overwrite persistent values with defaults
        if persistent:
            self.writeDefault = False
        self.inst = inst
        if hasattr(self.inst, "_api"):
            self.set = self._set_pynetworktables
        else:
            self.set = self._set_pyntcore

        self.reset()

    def reset(self):
        self.ntvalue = self.inst.getGlobalAutoUpdateValue(
            self.key, self.defaultValue, self.writeDefault
        )
        if self.persistent:
            self.ntvalue.setPersistent()

        # this is an optimization, but presumes the value type never changes
        self.mkv = Value.getFactoryByType(self.ntvalue.getType())

    def get(self, _):
        return self.ntvalue.value

    def _set_pynetworktables(self, _, value):
        self.inst._api.setEntryValueById(self.ntvalue._local_id, self.mkv(value))

    def _set_pyntcore(self, _, value):
        self.ntvalue.setValue(self.mkv(value))


def ntproperty(
    key: str,
    defaultValue,
    writeDefault: bool = True,
    doc: str = None,
    persistent: bool = False,
    *,
    inst: NetworkTablesInstance = NetworkTables
) -> property:
    """
        A property that you can add to your classes to access NetworkTables
        variables like a normal variable.
        
        :param key: A full NetworkTables key (eg ``/SmartDashboard/foo``)
        :param defaultValue: Default value to use if not in the table
        :type  defaultValue: any
        :param writeDefault: If True, put the default value to the table,
                             overwriting existing values
        :param doc: If given, will be the docstring of the property.
        :param persistent: If True, persist set values across restarts.
                           *writeDefault* is ignored if this is True.
        :param inst: The NetworkTables instance to use.

        Example usage::
        
            class Foo(object):
            
                something = ntproperty('/SmartDashboard/something', True)
                
                ...
                
                def do_thing(self):
                    if self.something:    # reads from value
                        ...
                        
                        self.something = False # writes value
        
        .. note:: Does not work with empty lists/tuples.
        
                  Getting the value of this property should be reasonably
                  fast, but setting the value will have just as much overhead
                  as :meth:`.NetworkTable.putValue`

        .. warning::

           This function assumes that the value's type
           never changes. If it does, you'll get really strange
           errors... so don't do that.

        .. versionadded:: 2015.3.0

        .. versionchanged:: 2017.0.6
            The *doc* parameter.

        .. versionchanged:: 2018.0.0
            The *persistent* parameter.
    """
    ntprop = _NtProperty(key, defaultValue, writeDefault, persistent, inst)
    try:
        inst._ntproperties.add(ntprop)
    except AttributeError:
        pass  # pyntcore compat

    return property(fget=ntprop.get, fset=ntprop.set, doc=doc)


class ChooserControl(object):
    """
        Interacts with a :class:`wpilib.SendableChooser`
        object over NetworkTables.
    """

    def __init__(
        self,
        key: str,
        on_choices: Optional[Callable[[Sequence[str]], None]] = None,
        on_selected: Optional[Callable[[str], None]] = None,
        *,
        inst: NetworkTablesInstance = NetworkTables
    ) -> None:
        """
            :param key: NetworkTables key
            :param on_choices: A function that will be called when the
                               choices change.
            :param on_selection: A function that will be called when the
                                 selection changes.
            :param inst: The NetworkTables instance to use.
        """
        self.subtable = inst.getTable("SmartDashboard").getSubTable(key)

        self.on_choices = on_choices
        self.on_selected = on_selected

        if on_choices or on_selected:
            self.subtable.addTableListener(self._on_change, True)

    def close(self) -> None:
        """Stops listening for changes to the ``SendableChooser``"""
        if self.on_choices or self.on_selected:
            self.subtable.removeTableListener(self._on_change)

    def getChoices(self) -> Sequence[str]:
        """
            Returns the current choices. If the chooser doesn't exist, this
            will return an empty tuple.
        """
        return self.subtable.getStringArray("options", ())

    def getSelected(self) -> Optional[str]:
        """
            Returns the current selection or None
        """
        selected = self.subtable.getString("selected", None)
        if selected is None:
            selected = self.subtable.getString("default", None)
        return selected

    def setSelected(self, selection: str) -> None:
        """
            Sets the active selection on the chooser
            
            :param selection: Active selection name
        """
        self.subtable.putString("selected", selection)

    def _on_change(self, table, key, value, isNew):
        if key == "options":
            if self.on_choices is not None:
                self.on_choices(value)
        elif key == "selected":
            if self.on_selected is not None:
                self.on_selected(value)
        elif key == "default":
            if (
                self.on_selected is not None
                and self.subtable.getString("selected", None) is None
            ):
                self.on_selected(value)



#!/usr/bin/env python3
#
# Simple benchmark script that tests NetworkTables throughput.
#
# To run the server
#
#     python3 benchmark.py
#
# To run the client:
#
#    python3 benchmark.py 127.0.0.1
#
# In theory, the limiting factor is internal buffering and not network
# bandwidth, so running this test on a local machine should be mostly
# valid.
#
# On OSX with default write flush period, I average ~18hz
#

from __future__ import print_function

from argparse import ArgumentParser
import time

import logging

logging.basicConfig(level=logging.DEBUG)

from networktables import NetworkTables


class Benchmark(object):
    def __init__(self):

        parser = ArgumentParser()
        parser.add_argument("client", nargs="?", default=None)
        parser.add_argument("-r", "--rate", default=None, type=float)
        parser.add_argument("--send", default=False, action="store_true")

        args = parser.parse_args()

        if args.client:
            NetworkTables.initialize(server=args.client)
        else:
            NetworkTables.initialize()

        # Default write flush is 0.05, could adjust for less latency
        if args.rate is not None:
            print("Setting rate to %s" % args.rate)
            NetworkTables.setUpdateRate(args.rate)

        self.nt = NetworkTables.getTable("/benchmark")
        self.updates = 0

        self.nt.addTableListener(self.on_update)

        if args.send:
            self.send_benchmark()
        else:
            self.recv_benchmark()

    def on_update(self, *args):
        self.updates += 1

    def recv_benchmark(self):

        print("Starting to receive")

        last = None
        last_updates = 0

        while True:

            now = time.time()
            updates = self.updates

            if last is not None:
                rate = (updates - last_updates) / (now - last)
                print("Update rate:", rate)

            last = now
            last_updates = updates
            time.sleep(1)

    def send_benchmark(self):

        print("Sending")

        i = 0

        while True:
            i += 1
            self.nt.putNumber("key", i)
            time.sleep(0.0001)


if __name__ == "__main__":
    Benchmark()
